/*
 * Perform SPI communication using the SPI peripheral.
 *
 * Copyright (C) 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/re/c/dma_space.h>
#include <l4/sys/types.h>
#include <stdint.h>



#ifdef __cplusplus

#include <l4/devices/cpm-generic.h>
#include <l4/devices/dma-generic.h>
#include <l4/devices/hw_mmio_register_block.h>
#include <l4/devices/spi.h>

/* SPI peripheral channel. */

class Spi_channel : public Spi_channel_base
{
  Hw::Register_block<32> _regs;

  /* Initialisation parameters. */

  l4_addr_t _spi_start;
  enum Clock_identifiers _clock;
  Cpm_chip *_cpm;
  Dma_channel *_dma;
  int _request_type;
  uint64_t _frequency;

  /* Common utilities. */

  void configure_transfer(uint8_t char_size);
  void wait_busy();

public:
  explicit Spi_channel(l4_addr_t spi_start, l4_addr_t start,
                       enum Clock_identifiers clock,
                       Cpm_chip *cpm,
                       Dma_channel *dma,
                       int request_type,
                       uint64_t frequency);

  /* Convenience operations. */

  virtual uint32_t send(uint32_t bytes, const uint8_t data[]);

  virtual uint32_t send_dc(uint32_t bytes, const uint8_t data[], const int dc[],
                           uint8_t char_size, bool big_endian = true);

  uint32_t send_units(uint32_t bytes, const uint8_t data[], uint8_t unit_size,
                      uint8_t char_size, bool big_endian = true);

  /* DMA operations. */

  uint32_t transfer(l4_addr_t vaddr, l4re_dma_space_dma_addr_t paddr,
                    uint32_t count, uint8_t unit_size, uint8_t char_size,
                    l4_addr_t desc_vaddr = 0,
                    l4re_dma_space_dma_addr_t desc_paddr = 0);
};

/* SPI peripheral. */

class Spi_chip
{
protected:
  l4_addr_t _spi_start, _start, _end;
  Cpm_chip *_cpm;

  virtual unsigned int num_channels() = 0;

  virtual Spi_channel *_get_channel(uint8_t channel, Dma_channel *dma,
                                    uint64_t frequency) = 0;

public:
  explicit Spi_chip(l4_addr_t spi_start, l4_addr_t start, l4_addr_t end, Cpm_chip *cpm);

  Spi_channel *get_channel(uint8_t channel, Dma_channel *dma, uint64_t frequency);
};

#endif /* __cplusplus */
